﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.UI;
using System.Xml;
using System.Xml.Serialization;
using System.IO;
using System;

public class Entity : MonoBehaviour {

	[Tooltip("The id of the entity, used to get the current stage from the database.")]
	public int entityId;

	[Space(10)][Tooltip("Can the player interact with the entity ?")]
	public bool canBeInteractedWith;

	
	[Space(10)][Tooltip("List of stages (progression levels) of the entity.")]
	public List<stages> stages = new List<stages>();

	[Space(10)][Tooltip("A list of all displayable options.")]
	public List<options> options = new List<options>();

	[Space(10)][Tooltip("A list of all nodes (event points).")]
	public List<nodes> nodes = new List<nodes>();

	//xml options
	public string XMLImportPath;
	public string XMLExportPath;

	public string XMLImportFileName;
	public string XMLExportFileName;

	[HideInInspector]
	public int tab = 0;
	
	//responsible for processing and displaying dialogue and options at the start
	private void dialogueInitiator () {
		//getting current stage
		int currentStageId = Database.core.entities[SupportFunctions.core.findEntityIndexById(entityId)].entityStageId;
		int stageIndex = findStageIndexByStageId(currentStageId);

		bool allowed = true;
		if ((stageIndex == -1) || !canBeInteractedWith) {
			allowed = false;
		}

		if (allowed) {
			//initiating dialogue
			dialogueEngine (0, stageIndex, -1, -1);
		}

	}

	public void dialogueEngine (int displayType, int stageIndex, int optionId, int nodeId) {

		if (displayType == 0) {
			//Display type 0 should be solely used at the start of an interaction to initiate a dialogue / action sequence
			//recursively calling the function with node data
			dialogueEngine (1, stageIndex, -1, stages[stageIndex].startingNodeId);

		} else if (displayType == 1) {

			ReferenceHolder.core.ClickCatcher.SetActive(true);

			//Display a node
			//index
			var nodeIndex = findNodeIndexById (nodeId);
			var allowed = true;

			//checking conditions before proceeding
			if (nodes[nodeIndex].conditionsToCheck.Count > 0) {
				foreach (condition c in nodes[nodeIndex].conditionsToCheck) {
					if (c.checkItemAcquired) {
						var itemIndex = SupportFunctions.core.findItemIndexById(c.acquiredItemId);

						if (!Database.core.items[itemIndex].acquired) {
							allowed = false;
							dialogueEngine (1, stageIndex, -1, c.redirectNodeIfAcquiredFails);
							break;
						}
					}

					if (c.checkItemQuantity) {

						var itemIndex = SupportFunctions.core.findItemIndexById(c.quantityItemId);

						if (!Database.core.items[itemIndex].acquired) {
							dialogueEngine (1, stageIndex, -1, c.redirectNodeIfQuantityFails);
							allowed = false;
							break;
						} else {
							if (Database.core.items[itemIndex].quantity < c.requiredQuantity) {
								allowed = false;
								dialogueEngine (1, stageIndex, -1, c.redirectNodeIfQuantityFails);
								break;
							}
						}
					}

					if (c.checkQuestExists) {
						var questIndex = SupportFunctions.core.findQuestIndexById(c.questExistsId);

						if (!Database.core.quests[questIndex].questAcquired) {
							allowed = false;
							dialogueEngine (1, stageIndex, -1, c.redirectNodeIfExistsFails);
							break;
						}
					}

					if (c.checkQuestCompleted) {
						var questIndex = SupportFunctions.core.findQuestIndexById(c.questCompletedId);

						if (!Database.core.quests[questIndex].questAcquired) {
							allowed = false;
							dialogueEngine (1, stageIndex, -1, c.redirectNodeIfCompletedFails);
							break;
						} else {
							if (!Database.core.quests[questIndex].questCompleted) {
								allowed = false;
								dialogueEngine (1, stageIndex, -1, c.redirectNodeIfCompletedFails);
								break;
							}
						}
					}
				}
			}
			
			if (allowed) {

				//display text?
				if (nodes[nodeIndex].displayText) {
					//if so, call special function
					Speech.core.genText (nodes[nodeIndex], stageIndex, this.gameObject);
					//add to journal
					addToJournal(nodes[nodeIndex].text, nodes[nodeIndex]);

					if (nodes[nodeIndex].goToNodeOnClick) {
						Button clickCatcher = ReferenceHolder.core.ClickCatcher.GetComponent<Button>();
						clickCatcher.onClick.AddListener( delegate { clickCatcher.onClick.RemoveAllListeners(); dialogueEngine(1, stageIndex, -1, nodes[nodeIndex].goToNodeId); } );
					}

				} else {
					
					if (nodes[nodeIndex].goToNodeOnClick) {
						dialogueEngine(1, stageIndex, -1, nodes[nodeIndex].goToNodeId);
					}
					
				}
				
				if (nodes[nodeIndex].displayOptions) {
					//making node click catcher non interactable
					ReferenceHolder.core.ClickCatcher.SetActive(false);
					optionSpawner (nodes[nodeIndex].optionsIds);
				}

				if (nodes[nodeIndex].setNewStageId) {
					Database.core.entities[SupportFunctions.core.findEntityIndexById(entityId)].entityStageId = nodes[nodeIndex].newStageId;
				}

				//adding items
				if (nodes[nodeIndex].itemsToAdd.Count > 0) {
					foreach (addItems item in nodes[nodeIndex].itemsToAdd) {
						var itemIndex = SupportFunctions.core.findItemIndexById (item.itemId);

						if (itemIndex != -1) {
							Database.core.items[itemIndex].acquired = true;
							Database.core.items[itemIndex].quantity += item.quantityToAdd;
						}
					}
				}

				//removing items
				if (nodes[nodeIndex].itemsToRemove.Count > 0) {
					foreach (removeItems item in nodes[nodeIndex].itemsToRemove) {
						var itemIndex = SupportFunctions.core.findItemIndexById (item.itemId);

						if (itemIndex != -1) {
							if (Database.core.items[itemIndex].acquired) {
								if (item.removeAll) {
									Database.core.items[itemIndex].quantity = 0;
									Database.core.items[itemIndex].acquired = false;
								} else if (item.removeQuantity) {
									if ((Database.core.items[itemIndex].quantity - item.quantityToRemove) < 0) {
										Database.core.items[itemIndex].quantity = 0;
										Database.core.items[itemIndex].acquired = false;
									} else {
										Database.core.items[itemIndex].quantity -= item.quantityToRemove;
									}
								}
							}
							
						}
					}
				}

				//quests to get
				if (nodes[nodeIndex].questsToGet.Count > 0) {
					foreach (int quest in nodes[nodeIndex].questsToGet) {
						var questIndex = SupportFunctions.core.findQuestIndexById(quest);

						if (questIndex != -1) {
							//making sure that the quest has not already been completed. Remove this check to allow repeating quests
							if (!Database.core.quests[questIndex].questCompleted) {
								//getting the quest
								Database.core.quests[questIndex].questAcquired = true;
							}
							
						}
					}
				}

				//quests to complete
				if (nodes[nodeIndex].questsToComplete.Count > 0) {
					foreach (int quest in nodes[nodeIndex].questsToComplete) {
						var questIndex = SupportFunctions.core.findQuestIndexById(quest);

						if (questIndex != -1) {
							Database.core.quests[questIndex].questCompleted = true;
						}
					}
				}

				//custom events to invoke
				if (nodes[nodeIndex].invokeCustomEvents) {
					foreach (string e in nodes[nodeIndex].eventsToBeInvoked) {
						EntityCustomEvents.core.Invoke(e, 0f);
					}
				}

				//was there any text displayed?
				if (nodes[nodeIndex].displayText) {
					//if so, we add a listener to the current text area to end interacction on click
					if (nodes[nodeIndex].endDialogueOnClick) {
						Button clickCatcher = ReferenceHolder.core.ClickCatcher.GetComponent<Button>();
						clickCatcher.onClick.AddListener( delegate { clickCatcher.onClick.RemoveAllListeners(); endInteraction(); } );
					}
				} else {
					//otherwise we simply end the dialogue
					if (nodes[nodeIndex].endDialogueOnClick) {
						endInteraction();
					}
				}

				//load level
				if (nodes[nodeIndex].loadLevel) {
					LevelLoader.core.spawnCoordinates = new Vector2 (nodes[nodeIndex].loadSpawnX, nodes[nodeIndex].loadSpawnY);
					LevelLoader.core.load (nodes[nodeIndex].loadLevelId, false, null);
				}
			}
			

		} else {

			var optionIndex = findOptionIndexById (optionId);

			//Display an option
			if (options[optionIndex].goToNodeOnClick) {
				dialogueEngine(1, stageIndex, -1, options[optionIndex].goToNodeId);
			}

		}
		
	}

	//option spawner
	private void optionSpawner (List<int> optionIds) {
		//nodes will be spawned up to the spawn points limit
		//Getting spawn point object
		GameObject parent = ReferenceHolder.core.OptionSpawnPointParent;

		//destroying all the previously generated options
		foreach (Transform child in parent.transform) {
			if (child.childCount > 0) {
				Destroy (child.GetChild(0).gameObject);
			}
		}

		//getting spawn points and storing them in a list
		List<GameObject> spawnPoints = new List<GameObject>();

		foreach (Transform child in parent.transform) {
			spawnPoints.Add(child.gameObject);
		}

		var counter = 0;
		foreach (int i in optionIds) {
			if (spawnPoints.Count > counter) {
				GameObject instance = Instantiate(ReferenceHolder.core.OptionPrefab, spawnPoints[counter].transform);

				var optionIndex = findOptionIndexById(i);

				if (optionIndex != -1) {
					//change option text
					instance.transform.GetChild(0).gameObject.GetComponent<Text>().text = options[optionIndex].optionText;
					//add listener
					instance.GetComponent<Button>().onClick.AddListener(delegate { optionListener(i); });
					counter++;
				} else {
					Debug.Log("Invalid ID");
					break;
				}
				
			}
		}
		//enabling object
		if (counter > 0) {
			//if at least one option was generated
			ReferenceHolder.core.OptionObject.SetActive(true);
		}
	}

	private void optionListener (int opIndex) {
		ReferenceHolder.core.OptionObject.SetActive(false);
		//marking option as selected
		options[opIndex].selected = true;
		dialogueEngine (2, -1, opIndex, -1);
	}

	public void interact () {

		//once the interaction begins, all other actions should be blocked
		EntityInteractionManager.core.playerCanInteract = false;
		ItemInteractionManager.core.playerCanInteract = false;
		Movement.core.playerCanMove = false;
		ReferenceHolder.core.ClickCatcher.SetActive(true);

		dialogueInitiator ();

	}

	public void endInteraction () {
		EntityInteractionManager.core.playerCanInteract = true;
		ItemInteractionManager.core.playerCanInteract = true;
		Movement.core.playerCanMove = true;
		ReferenceHolder.core.ClickCatcher.SetActive(false);
		ReferenceHolder.core.OptionObject.SetActive(false);
	}

	//Checking if our mouse is hovering over the gameobject
	//we are using update instead of mouse over since we need to exclude certain layers from the "equation".
	private bool sw = false;

	void Update() {
		
		Vector2 point = Camera.main.ScreenToWorldPoint(Input.mousePosition);
		RaycastHit2D h = Physics2D.Raycast(point, Vector2.zero, Mathf.Infinity, ~(1 << LayerMask.NameToLayer("MovementCatcher")));

		if (h && h.collider.gameObject == this.gameObject) {

			sw = true;

			var entityIndex = SupportFunctions.core.findEntityIndexById(entityId);
			if (entityIndex != -1) {
				//name
				SupportFunctions.core.displayName(Database.core.entities[entityIndex].entityName, false);
				//cursor
				CursorChanger.core.switchCursor (false);

			} else {
				Debug.Log ("Wrong item id " + entityId.ToString());
			}
		
		} else if (sw) {
			sw = false;
			//Name
			SupportFunctions.core.displayName(null, true);
			//cursor
			CursorChanger.core.switchCursor (true);
		}
		

	}

	//adding entries to journals
	void addToJournal (string text, nodes node) {

		var name = PlayerManager.core.playerName;

		//if the player talking ?
		if (!node.playerIsTalking) {
			name = Database.core.entities[SupportFunctions.core.findEntityIndexById(node.talkingEntityId)].entityName;
		}
		
		//creating class instance
		journal entry = new journal();
		//assigning values
		entry.name = name;
		entry.entry = text;
		//adding to database
		Database.core.journal.Add(entry);

	}

	//getting stage list index by stage id
	public int findStageIndexByStageId (int id) {

		var counter = 0;

		foreach (stages stage in stages) {
			if (stage.stageId == id) {
				return counter;
			} else {
				counter++;
			}
		}
		
		return -1;
	}

	//find option index by id
	int findOptionIndexById (int id) {

		var counter = 0;

		foreach (options option in options) {
			if (option.optionId == id) {
				return counter;
			} else {
				counter++;
			}
		}
		
		return -1;
	}

	//find node index by id
	int findNodeIndexById (int id) {

		var counter = 0;

		foreach (nodes node in nodes) {
			if (node.nodeId == id) {
				return counter;
			} else {
				counter++;
			}
		}
		
		return -1;
	}

	//xml
	public void XMLImport () {
		import(XMLImportPath, XMLImportFileName);
	}

	public void XMLExport () {
		export(XMLExportPath, XMLExportFileName);
	}

	public void import (string path, string name) {

		try {

			nodes = ReadFromXmlFile <List<nodes>> (path + "\\" + name + ".xml");

			Debug.Log("File loaded from : " + path);

		} catch (Exception e) {
			Debug.Log("Oops, it seems that something went wrong. Error : " + e);
		} 

	}

	public void export (string path, string name) {

		List<nodes> n = nodes;

		try {

			WriteToXmlFile (path + "\\" + name + ".xml", n);

			Debug.Log("File created in : " + path);

		} catch (Exception e) {
			Debug.Log("Oops, it seems that something went wrong. Error : " + e);
		}

	}

	public static void WriteToXmlFile<T>(string filePath, T objectToWrite, bool append = false) where T : new() {
		
		TextWriter writer = null;

		try {
			var serializer = new XmlSerializer(typeof(T));
			writer = new StreamWriter(filePath, append);
			serializer.Serialize(writer, objectToWrite);
		}

		finally {

			if (writer != null) {
				writer.Close();
			}
		}
	}

	public static T ReadFromXmlFile<T>(string filePath) where T : new() {
		
		TextReader reader = null;

		try {
			var serializer = new XmlSerializer(typeof(T));
			reader = new StreamReader(filePath);
			return (T)serializer.Deserialize(reader);
		}

		finally {
			
			if (reader != null) {
				reader.Close();
			}
			
		}
	}

}

[System.Serializable]
public class stages {

	[Tooltip("Stage id.")]
	public int stageId;

	[Space(10)][Tooltip("The id of the node to be accessed upon entering the stage.")]
	public int startingNodeId;
}

[System.Serializable]
public class nodes {

	[Tooltip("Name of element in the list.")][XmlAttribute("elementName")]
	public string ElementName;

	[Space(10)][Tooltip("A unique node id.")][XmlAttribute("nodeID")]
	public int nodeId;

	[Space(10)][Tooltip("Should text be displayed ?")][XmlAttribute("displayText")]
	public bool displayText;

	[Space(10)][Tooltip("The coordinates of this gameObject will be used to spawn the speechBubble text prefab. Not required if player is talking.")][XmlIgnore]
	public GameObject textTarget;

	[Space(10)][Tooltip("The text to be displayed.")][XmlAttribute("text")]
	public string text;

	[Space(10)][Tooltip("How long should the text be displayed for ?")][XmlAttribute("textDuration")]
	public float textDuration;

	[Space(10)][Tooltip("The text will be displayed until the player decides to skip it.")][XmlAttribute("waitForClick")]
	public bool waitForClick;

	[Space(10)][Tooltip("Is player the one doing the talking ?")][XmlAttribute("playerIsTalking")]
	public bool playerIsTalking;

	[Space(10)][Tooltip("If the player is not the one talking, what is the id of the entity which is ?")][XmlAttribute("talkingEntityId")]
	public int talkingEntityId;

	//Custom events
	[Space(10)][Tooltip("Invoke custom functions from EntityCustomEvents.cs ?")][XmlAttribute("invokeCustomEvents")]
	public bool invokeCustomEvents;

	[Space(10)][Tooltip("A list of function names from EntityCustomEvents.cs to be invoked.")][XmlIgnore]
	public List<string> eventsToBeInvoked = new List<string>();

	//Audio
	[Space(10)][Tooltip("Play audio along with text ?")][XmlAttribute("playAudioWithText")]
	public bool playAudioWithText;

	[Space(10)][Tooltip("AudioClip to be played.")][XmlIgnore]
	public AudioClip audio;

	//check condition before proceeding
	[Space(10)][Tooltip("A list of conditions to be checked before allowing any other events.")][XmlIgnore]
	public List<condition> conditionsToCheck = new List<condition>();

	//next node
	[Space(10)][Tooltip("The next node to be accessed. If there was no dialogue displayed beforehand, the next node will be loaded immediately.")][XmlAttribute("goToNode")]
	public bool goToNodeOnClick;

	[Space(10)][Tooltip("The id of the next node to be loaded.")][XmlAttribute("goToNodeId")]
	public int goToNodeId;

	//options
	[Space(10)][Tooltip("Should options be displayed ?")][XmlAttribute("displayOptions")]
	public bool displayOptions;

	[Space(10)][Tooltip("A list of options (their ids) to be displayed. Please make sure that the amount of options to be displayed does not exceed the option spawn points amount.")]
	[XmlIgnore]
	public List<int> optionsIds = new List<int>();

	//set stage id
	[Space(10)][Tooltip("A new entity stage id ?")][XmlAttribute("setNewStageId")]
	public bool setNewStageId;

	[Space(10)][Tooltip("Id to be set.")][XmlAttribute("newStageId")]
	public int newStageId;

	//inventory
	[Space(10)][Tooltip("A list of item ids to add to the inventory.")][XmlIgnore]
	public List<addItems> itemsToAdd = new List<addItems>();
	
	[Space(10)][Tooltip("Item ids to be removed.")][XmlIgnore]
	public List<removeItems> itemsToRemove = new List<removeItems>();

	//QUESTS
	[Space(10)][Tooltip("Ids of quests to get.")][XmlIgnore]
	public List<int> questsToGet = new List<int> ();
	
	[Space(10)][Tooltip("Ids of quests to be completed.")][XmlIgnore]
	public List<int> questsToComplete = new List<int>();

	[Space(10)][Tooltip("End dialogue on click. If there was no dialogue previously shown, the dialogue will end immediately.")][XmlAttribute("endDialogue")]
	public bool endDialogueOnClick;

	//loading
	[Space(10)][Tooltip("Load Level ?")][XmlAttribute("loadLevel")]
	public bool loadLevel;

	[Space(10)][Tooltip("ID of the level to load in the database.")][XmlAttribute("loadLevelId")]
	public int loadLevelId;

	[Space(10)][Tooltip("The X coordinate on which the player will spawn after loading.")][XmlAttribute("loadSpawnX")]
	public float loadSpawnX;

	[Space(10)][Tooltip("The Y coordinate on which the player will spawn after loading.")][XmlAttribute("loadSpawnY")]
	public float loadSpawnY;
}

[System.Serializable]
public class options {

	[Space(10)][Tooltip("A unique option id.")]
	public int optionId;

	[Space(10)][Tooltip("Option text.")]
	public string optionText;

	//go to node on click
	[Space(10)][Tooltip("Go to node on option click.")]
	public bool goToNodeOnClick;

	[Space(10)][Tooltip("Node id.")]
	public int goToNodeId;

	[HideInInspector]
	public bool selected;

}

[System.Serializable]
public class removeItems {

	[Tooltip("Id of item to remove.")]
	public int itemId;

	[Space(10)][Tooltip("Remove the item completely.")]
	public bool removeAll;

	[Space(10)][Tooltip("Remove a certain quantity of the item ?")]
	public bool removeQuantity;

	[Space(10)][Tooltip("Quantity to remove.")]
	public int quantityToRemove;
}

[System.Serializable]
public class addItems {
	[Tooltip("Item id.")]
	public int itemId;
	[Space(10)][Tooltip("Add x? of this item to inventory.")]
	public int quantityToAdd = 1;
}

[System.Serializable]
public class condition {

	//quest exists
	[Tooltip("Check if player has obtained a certain quest ?")]
	public bool checkQuestExists;

	[Space(10)][Tooltip("The id of the quest to be checked.")]
	public int questExistsId;

	[Space(10)][Tooltip("The id of the node to be accessed if check fails.")]
	public int redirectNodeIfExistsFails;

	//quest completed
	[Space(10)][Tooltip("Has the quest been completed ?")]
	public bool checkQuestCompleted;

	[Space(10)][Tooltip("The id of the quest to be checked.")]
	public int questCompletedId;

	[Space(10)][Tooltip("The id of the node to be accessed if check fails.")]
	public int redirectNodeIfCompletedFails;

	//item acquired
	[Space(10)][Tooltip("Check if player has a specific item.")]
	public bool checkItemAcquired;

	[Space(10)][Tooltip("Id of the item to be checked.")]
	public int acquiredItemId;

	[Space(10)][Tooltip("The id of the node to be accessed if check fails.")]
	public int redirectNodeIfAcquiredFails;

	//quantity
	[Space(10)][Tooltip("Check for a specific item quantity.")]
	public bool checkItemQuantity;
	
	[Space(10)][Tooltip("Id of item.")]
	public int quantityItemId;

	[Space(10)][Tooltip("Minimum amount of item required.")]
	public int requiredQuantity;

	[Space(10)][Tooltip("The id of the node to be accessed if check fails.")]
	public int redirectNodeIfQuantityFails;
}

//Tsenkilidis Alexandros - Cination