﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;

//character movement
public class Movement : MonoBehaviour {

	public static Movement core;

	//public
	[Tooltip("The time interval between each iteration of the position check.")]
	public float movementRefreshSpeed;

	[Tooltip("Play step sound ?")]
	public bool playStepSound;

	[Tooltip("Delay between step clip plays.")]
	public float stepClipSpeed;

	//private
	private AudioSource stepSource;

	public Coroutine move;

	Coroutine movementSoundRoutine;

	private GameObject player;

	private Animator animator;

	[Tooltip("How should the x, y spawn position of the raycast box be adjusted ?")]
	public Vector2 rayCastSpawnAdjustments = new Vector2(0,0);

	[Tooltip("RayCastBoxSize")]
	//RaycastHit2D hit;
	public Vector2 size = new Vector2(0.5f, 0.5f);

	public float castDistance = 0.5f;

	//This variable can be accessed by any "animationDirector.cs" script to indicate the current animation drection
	public string animationDirection = "";

	[Header ("Player animator trasnition parameters.")]
	public movementAnimations matp;

	[Space(10)][Tooltip("Should the player be able to move ?")]
	public bool playerCanMove = true;

	[HideInInspector]
	public bool moving;

	[HideInInspector]
	public Vector2 coordinates;


	public void Move () {

		if (move != null) {
			StopCoroutine(move);
		}

		if (movementSoundRoutine != null) {
			StopCoroutine(movementSoundRoutine);
		}

		player.GetComponent<CircleCollider2D>().enabled = true;
		move = StartCoroutine(IEm());
		if (playStepSound) {
			movementSoundRoutine = StartCoroutine(IEmS());
		}

	}

	public void StopMovement (bool onLoad) {

		//stopping movement
		if (move != null) {

			StopCoroutine(move);
			abm(matp.idle);

			player.GetComponent<CircleCollider2D>().enabled = false;

			moving = false;

		}
		
		StopSound();
	}

	void StopSound () {

		//stopping sound
		if (movementSoundRoutine != null) {
			StopCoroutine (movementSoundRoutine);
		}

		if (stepSource != null) {
			Destroy (stepSource);
		}

	}
	
	IEnumerator IEm () {

		moving = true;

		Vector2 playerV2 = player.transform.position;

		//animation
		Vector3 playerPos = player.transform.position;
		Vector3 modCoordinates = new Vector3 (coordinates.x, coordinates.y, player.transform.position.z);

		while (true) {

			yield return new WaitForSeconds (movementRefreshSpeed);
			playerPos = player.transform.position;
			
			if (playerPos == modCoordinates) {

				//Player has arrived to destination
				StopMovement(false);

				break;

			} else {

				bool allow = true;
				playerV2 = new Vector2 (player.transform.position.x + rayCastSpawnAdjustments.x, player.transform.position.y + rayCastSpawnAdjustments.y);

				/*Legacy
				var heading = coordinates - playerV2;
				var distance = heading.magnitude;
				var direction = heading / distance;

				var hit1 = Physics2D.BoxCast (playerV2, new Vector2(1f, 1f), 0, direction, 0, 1 << LayerMask.NameToLayer("collisionLayer"));
				var hit2 = Physics2D.BoxCast (playerV2, size, 0, direction, castDistance, 1 << LayerMask.NameToLayer("collisionLayer"));
				*/

				var hit = Physics2D.Linecast (playerV2, coordinates, 1 << LayerMask.NameToLayer("CollisionLayer"));
				
				if (hit) {

					Vector2 hitPos = (Vector2) hit.point;

					var colDif = hitPos - playerV2;

					//Debug.DrawLine (playerV2, coordinates, new Color(0.0f, 0.0f, 1.0f));

					if (Mathf.Abs(colDif.x) < 1 || Mathf.Abs (colDif.y) < 1) {
						allow = false;
					}

					
				}

				if (allow) {
					
					var step = PlayerManager.core.movementSpeed * Time.deltaTime;
					player.transform.position = Vector3.MoveTowards(playerPos, modCoordinates, step);
					PlayAnimation(animationDirection);
					
				} else {
					
					StopMovement(false);

				}
				
			}
		}
	}

	IEnumerator IEmS () {

		//getting audio gameObject
		var AudioHolder = ReferenceHolder.core.AudioHolder;
		//creating source
		stepSource = AudioHolder.AddComponent<AudioSource>();
		stepSource.playOnAwake = false;

		//to get the clip, let's find the clip id for this level
		var levelManagerClipId = GameObject.FindGameObjectWithTag("levelManager").GetComponent<LevelManager>().stepSoundId;

		stepSource.clip = Database.core.stepSounds[SupportFunctions.core.findStepClipIndexById(levelManagerClipId)].clip;

		while (true) {
			stepSource.Play();
			yield return new WaitForSeconds (stepClipSpeed);
			stepSource.Stop();
		}

	}

	void PlayAnimation(string dir) {

		//calling abm to play proper animation
		switch (dir) {
			case "n":
				abm (matp.north);
				break;
			case "s":
				abm (matp.south);
				break;
			case "e":
				abm (matp.east);
				break;
			case "w":
				abm (matp.west);
				break;
			case "se":
				abm (matp.southEast);
				break;
			case "nw":
				abm (matp.northWest);
				break;
			case "sw":
				abm (matp.southWest);
				break;
			case "ne":
				abm (matp.northWest);
				break;
			default:
				break;
		}
	}

	//set animator booleans
	void abm (string dir) {

		//array of animator parameters
		List<string> ls = new List<string>();
		
        Animator a = player.GetComponent<Animator>();

        foreach (AnimatorControllerParameter param in a.parameters) {
			ls.Add(param.name);
		}
		
		foreach (string s in ls) {
			if (dir == s) {
				animator.SetBool(s, true);	
			} else {
				animator.SetBool(s, false);
			}
		}

	}

	void Start() {

		//getting player object
		player = GameObject.FindGameObjectWithTag("Player");
		//getting anim.
		animator = player.GetComponent<Animator>();

		animationDirection = matp.idle;
	}

	void Awake() {
		if (core == null) {
			core = this;
		}
	}
}

[System.Serializable]
public class movementAnimations {
	public string north;
	public string south;
	public string east;
	public string west;
	public string southEast;
	public string northWest;
	public string southWest;
	public string northEast;
	public string idle;
}

//Tsenkilidis Alexandros - Cination